import Module from 'module';
import { dirname, resolve, relative, basename } from 'pathe';
import chalk from 'chalk';
import consola from 'consola';
import defu from 'defu';
import prettyBytes from 'pretty-bytes';
import jiti from 'jiti';
import mkdirp from 'mkdirp';
import fsp, { writeFile } from 'fs/promises';
import { promisify } from 'util';
import _rimraf from 'rimraf';
import { rollup } from 'rollup';
import commonjs from '@rollup/plugin-commonjs';
import { nodeResolve } from '@rollup/plugin-node-resolve';
import alias from '@rollup/plugin-alias';
import _esbuild from 'rollup-plugin-esbuild';
import dts from 'rollup-plugin-dts';
import rollupJSONPlugin from '@rollup/plugin-json';
import { createFilter } from '@rollup/pluginutils';
import { findStaticImports } from 'mlly';
import MagicString from 'magic-string';
import { resolveSchema, generateMarkdown, generateTypes } from 'untyped';
import untypedPlugin from 'untyped/loader/babel';
import { pascalCase } from 'scule';
import { mkdist } from 'mkdist';

async function ensuredir(path) {
  await mkdirp(dirname(path));
}
async function symlink(from, to, force = true) {
  await ensuredir(to);
  if (force) {
    await fsp.unlink(to).catch(() => {
    });
  }
  await fsp.symlink(from, to, "junction");
}
function dumpObject(obj) {
  return "{ " + Object.keys(obj).map((key) => `${key}: ${JSON.stringify(obj[key])}`).join(", ") + " }";
}
function getpkg(id = "") {
  const s = id.split("/");
  return s[0][0] === "@" ? `${s[0]}/${s[1]}` : s[0];
}
const rimraf = promisify(_rimraf);
async function rmdir(dir) {
  await fsp.unlink(dir).catch(() => {
  });
  await rimraf(dir);
}

function validateDependencies(ctx) {
  const usedDependencies = new Set();
  const unusedDependencies = new Set(Object.keys(ctx.pkg.dependencies || {}));
  const implicitDependnecies = new Set();
  for (const id of ctx.usedImports) {
    unusedDependencies.delete(id);
    usedDependencies.add(id);
  }
  if (Array.isArray(ctx.dependencies)) {
    for (const id of ctx.dependencies) {
      unusedDependencies.delete(id);
    }
  }
  for (const id of usedDependencies) {
    if (!ctx.externals.includes(id) && !id.startsWith("chunks/") && !ctx.dependencies.includes(getpkg(id))) {
      implicitDependnecies.add(id);
    }
  }
  if (unusedDependencies.size) {
    consola.warn("Potential unused dependencies found:", Array.from(unusedDependencies).map((id) => chalk.cyan(id)).join(", "));
  }
  if (implicitDependnecies.size && !ctx.inlineDependencies) {
    consola.warn("Potential implicit dependencies found:", Array.from(implicitDependnecies).map((id) => chalk.cyan(id)).join(", "));
  }
}

const EXPORT_DEFAULT = "export default ";
function JSONPlugin(options) {
  const plugin = rollupJSONPlugin(options);
  return {
    ...plugin,
    name: "unbuild-json",
    transform(code, id) {
      const res = plugin.transform.call(this, code, id);
      if (res && typeof res !== "string" && "code" in res && res.code && res.code.startsWith(EXPORT_DEFAULT)) {
        res.code = res.code.replace(EXPORT_DEFAULT, "module.exports = ");
      }
      return res;
    }
  };
}

const defaults = {
  include: [/\.(md|txt|css|htm|html)$/],
  exclude: []
};
function rawPlugin(opts = {}) {
  opts = { ...opts, ...defaults };
  const filter = createFilter(opts.include, opts.exclude);
  return {
    name: "unbuild-raw",
    transform(code, id) {
      if (filter(id)) {
        return {
          code: `export default ${JSON.stringify(code)}`,
          map: null
        };
      }
    }
  };
}

function cjsPlugin(_opts) {
  return {
    name: "unbuild-cjs",
    renderChunk(code, _chunk, opts) {
      if (opts.format === "es") {
        return CJSToESM(code);
      }
      return null;
    }
  };
}
const CJSyntaxRe = /__filename|__dirname|require\(|require\.resolve\(/;
const CJSShim = `

// -- Unbuild CommonJS Shims --
import __cjs_url__ from 'url';
import __cjs_path__ from 'path';
import __cjs_mod__ from 'module';
const __filename = __cjs_url__.fileURLToPath(import.meta.url);
const __dirname = __cjs_path__.dirname(__filename);
const require = __cjs_mod__.createRequire(import.meta.url);
`;
function CJSToESM(code) {
  if (code.includes(CJSShim) || !CJSyntaxRe.test(code)) {
    return null;
  }
  const lastESMImport = findStaticImports(code).pop();
  const indexToAppend = lastESMImport ? lastESMImport.end : 0;
  const s = new MagicString(code);
  s.appendRight(indexToAppend, CJSShim);
  return {
    code: s.toString(),
    map: s.generateMap()
  };
}

const esbuild = _esbuild.default || _esbuild;
async function rollupBuild(ctx) {
  if (ctx.stub) {
    for (const entry of ctx.entries.filter((entry2) => entry2.builder === "rollup")) {
      const output = resolve(ctx.rootDir, ctx.outDir, entry.name);
      if (ctx.emitCJS) {
        await writeFile(output + ".cjs", `module.exports = require('jiti')(null, { interopDefault: true })('${entry.input}')`);
      }
      await writeFile(output + ".mjs", `import jiti from 'jiti';
export default jiti(null, { interopDefault: true })('${entry.input}');`);
      await writeFile(output + ".d.ts", `export * from '${entry.input}'`);
    }
    return;
  }
  const rollupOptions = getRollupOptions(ctx);
  if (!Object.keys(rollupOptions.input).length) {
    return;
  }
  const buildResult = await rollup(rollupOptions);
  const allOutputOptions = rollupOptions.output;
  for (const outputOptions of allOutputOptions) {
    const { output } = await buildResult.write(outputOptions);
    for (const entry of output.filter((e) => e.type === "chunk")) {
      for (const id of entry.imports) {
        ctx.usedImports.add(id);
      }
      if (entry.isEntry) {
        ctx.buildEntries.push({
          path: relative(ctx.rootDir, resolve(outputOptions.dir, entry.fileName)),
          bytes: entry.code.length * 4,
          exports: entry.exports
        });
      }
    }
  }
  if (ctx.declaration) {
    rollupOptions.plugins = rollupOptions.plugins || [];
    rollupOptions.plugins.push(dts({
      respectExternal: true
    }));
    const typesBuild = await rollup(rollupOptions);
    await typesBuild.write({
      dir: resolve(ctx.rootDir, ctx.outDir),
      format: "esm"
    });
  }
}
function getRollupOptions(ctx) {
  const extensions = [".ts", ".mjs", ".cjs", ".js", ".json"];
  return {
    context: ctx.rootDir,
    input: Object.fromEntries(ctx.entries.filter((entry) => entry.builder === "rollup").map((entry) => [entry.name, resolve(ctx.rootDir, entry.input)])),
    output: [
      ctx.emitCJS && {
        dir: resolve(ctx.rootDir, ctx.outDir),
        entryFileNames: "[name].cjs",
        chunkFileNames: "chunks/[name].cjs",
        format: "cjs",
        exports: "auto",
        preferConst: true,
        externalLiveBindings: false,
        freeze: false
      },
      {
        dir: resolve(ctx.rootDir, ctx.outDir),
        entryFileNames: "[name].mjs",
        chunkFileNames: "chunks/[name].mjs",
        format: "esm",
        exports: "auto",
        preferConst: true,
        externalLiveBindings: false,
        freeze: false
      }
    ].filter(Boolean),
    external(id) {
      const pkg = getpkg(id);
      const isExplicitExternal = ctx.externals.includes(pkg);
      if (isExplicitExternal) {
        return true;
      }
      if (ctx.inlineDependencies || id[0] === "." || id[0] === "/" || id.match(/src[\\/]/) || id.startsWith(ctx.pkg.name)) {
        return false;
      }
      if (!isExplicitExternal) {
        consola.warn(`Inlining implicit external ${id}`);
      }
      return isExplicitExternal;
    },
    onwarn(warning, rollupWarn) {
      if (!warning.code || !["CIRCULAR_DEPENDENCY"].includes(warning.code)) {
        rollupWarn(warning);
      }
    },
    plugins: [
      alias({
        [ctx.pkg.name]: ctx.rootDir
      }),
      nodeResolve({
        extensions,
        preferBuiltins: true
      }),
      JSONPlugin({
        preferConst: true
      }),
      esbuild({
        target: "es2020"
      }),
      commonjs({
        extensions,
        ignoreTryCatch: true
      }),
      { renderDynamicImport() {
        return { left: "import(", right: ")" };
      } },
      ctx.cjsBridge && cjsPlugin(),
      rawPlugin()
    ].filter(Boolean)
  };
}

async function typesBuild(ctx) {
  for (const entry of ctx.entries.filter((entry2) => entry2.builder === "untyped")) {
    const _require = jiti(ctx.rootDir, {
      interopDefault: true,
      transformOptions: {
        babel: {
          plugins: [
            untypedPlugin
          ]
        }
      }
    });
    const distDir = entry.outDir;
    const srcConfig = _require(resolve(ctx.rootDir, entry.input));
    const defaults = entry.defaults || {};
    const schema = resolveSchema(srcConfig, defaults);
    await writeFile(resolve(distDir, `${entry.name}.md`), generateMarkdown(schema));
    await writeFile(resolve(distDir, `${entry.name}.schema.json`), JSON.stringify(schema, null, 2));
    await writeFile(resolve(distDir, `${entry.name}.defaults.json`), JSON.stringify(defaults, null, 2));
    if (entry.declaration) {
      await writeFile(resolve(distDir, `${entry.name}.d.ts`), "export " + generateTypes(schema, pascalCase(entry.name + "-schema")));
    }
  }
}

async function mkdistBuild(ctx) {
  for (const entry of ctx.entries.filter((e) => e.builder === "mkdist")) {
    const distDir = entry.outDir;
    if (ctx.stub) {
      await rmdir(distDir);
      await symlink(entry.input, distDir);
    } else {
      const { writtenFiles } = await mkdist({
        rootDir: ctx.rootDir,
        srcDir: entry.input,
        distDir,
        format: entry.format,
        cleanDist: false,
        declaration: entry.declaration,
        ext: entry.ext
      });
      ctx.buildEntries.push({
        path: distDir,
        chunks: [`${writtenFiles.length} files`]
      });
    }
  }
}

async function build(rootDir, stub) {
  rootDir = resolve(process.cwd(), rootDir || ".");
  const _require = jiti(rootDir);
  let buildConfigFile;
  try {
    buildConfigFile = _require.resolve("./build.config");
  } catch (e) {
  }
  const buildConfig = buildConfigFile ? _require("./build.config").default : {};
  const pkg = _require("./package.json");
  const ctx = defu(buildConfig, {
    pkg,
    rootDir,
    entries: [],
    dependencies: [],
    devDependencies: [],
    externals: [...Module.builtinModules],
    outDir: "dist",
    genDir: ".gen",
    untyped: void 0,
    declaration: void 0,
    inlineDependencies: false,
    clean: true,
    stub,
    buildEntries: [],
    usedImports: new Set(),
    emitCJS: true,
    cjsBridge: false
  });
  ctx.entries = ctx.entries.map((entry) => typeof entry === "string" ? { input: entry } : entry);
  for (const entry of ctx.entries) {
    if (typeof entry.name !== "string") {
      entry.name = basename(entry.input);
    }
    if (!entry.input) {
      throw new Error("Missing entry input: " + dumpObject(entry));
    }
    if (!entry.builder) {
      entry.builder = entry.input.endsWith("/") ? "mkdist" : "rollup";
    }
    if (ctx.declaration !== void 0 && entry.declaration === void 0) {
      entry.declaration = ctx.declaration;
    }
    entry.input = resolve(ctx.rootDir, entry.input);
    entry.outDir = resolve(ctx.rootDir, entry.outDir || ctx.outDir);
  }
  ctx.dependencies = Object.keys(pkg.dependencies || {});
  ctx.devDependencies = Object.keys(pkg.devDependencies || {});
  ctx.externals.push(...ctx.dependencies);
  consola.info(chalk.cyan(`${ctx.stub ? "Stubbing" : "Building"} ${pkg.name}`));
  if (process.env.DEBUG) {
    consola.info(`${chalk.bold("Root dir:")} ${ctx.rootDir}
  ${chalk.bold("Entries:")}
  ${ctx.entries.map((entry) => "  " + dumpObject(entry)).join("\n  ")}
`);
  }
  if (ctx.clean) {
    for (const dir of new Set(ctx.entries.map((e) => e.outDir).sort())) {
      await rmdir(dir);
      await mkdirp(dir);
    }
  }
  await typesBuild(ctx);
  await mkdistBuild(ctx);
  await rollupBuild(ctx);
  if (ctx.stub) {
    return;
  }
  consola.success(chalk.green("Build succeed for " + pkg.name));
  for (const entry of ctx.buildEntries) {
    consola.log(`  ${chalk.bold(entry.path)} (` + [
      entry.bytes && `size: ${chalk.cyan(prettyBytes(entry.bytes))}`,
      entry.exports && `exports: ${chalk.gray(entry.exports.join(", "))}`,
      entry.chunks && `chunks: ${chalk.gray(entry.chunks.join(", "))}`
    ].filter(Boolean).join(", ") + ")");
  }
  validateDependencies(ctx);
  consola.log("");
}

export { build as b };
